# Copyright (c) 2016-2022 Adam Karpierz
# Licensed under the zlib/libpng License
# https://opensource.org/licenses/Zlib

import unittest

import libusb


class MainTestCase(unittest.TestCase):

    def setUp(self):
        pass
