# Copyright (c) 2016-2022 Adam Karpierz
# Licensed under the zlib/libpng License
# https://opensource.org/licenses/Zlib

__import__("pkg_about").about()
__copyright__ = f"Copyright (c) 2016-2022 {__author__}"  # noqa
